@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2018-2020 Alexander Borisov
+ * Copyright (C) 2018-2022 Alexander Borisov
  *
  * Author: Alexander Borisov <borisov@lexbor.com>
  */
@@ -211,7 +211,7 @@ lxb_html_tree_token_callback(lxb_html_tokenizer_t *tkz,
     return token;
 }
 
-/* TODO: not complite!!! */
+/* TODO: not complete!!! */
 lxb_status_t
 lxb_html_tree_stop_parsing(lxb_html_tree_t *tree)
 {
@@ -459,8 +459,11 @@ lxb_html_tree_append_attributes(lxb_html_tree_t *tree,
 {
     lxb_status_t status;
     lxb_dom_attr_t *attr;
+    lxb_html_document_t *doc;
     lxb_html_token_attr_t *token_attr = token->attr_first;
 
+    doc = lxb_html_interface_document(element->node.owner_document);
+
     while (token_attr != NULL) {
         attr = lxb_dom_element_attr_by_local_name_data(element,
                                                        token_attr->name);
@@ -469,7 +472,7 @@ lxb_html_tree_append_attributes(lxb_html_tree_t *tree,
             continue;
         }
 
-        attr = lxb_dom_attr_interface_create(element->node.owner_document);
+        attr = lxb_dom_attr_interface_create(lxb_dom_interface_document(doc));
         if (attr == NULL) {
             return LXB_STATUS_ERROR_MEMORY_ALLOCATION;
         }
@@ -608,8 +611,8 @@ lxb_html_tree_adjust_foreign_attributes(lxb_html_tree_t *tree,
                                         lxb_dom_attr_t *attr, void *ctx)
 {
     size_t lname_length;
-    lexbor_hash_t *tags, *attrs, *prefix;
-    const lxb_tag_data_t *tag_data;
+    lexbor_hash_t *attrs, *prefix;
+    const lxb_dom_attr_data_t *attr_data;
     const lxb_ns_prefix_data_t *prefix_data;
     const lxb_dom_attr_data_t *data;
     const lxb_html_tree_res_attr_adjust_foreign_t *adjust;
@@ -617,7 +620,6 @@ lxb_html_tree_adjust_foreign_attributes(lxb_html_tree_t *tree,
     size_t len = sizeof(lxb_html_tree_res_attr_adjust_foreign_map)
         / sizeof(lxb_html_tree_res_attr_adjust_foreign_t);
 
-    tags = attr->node.owner_document->tags;
     attrs = attr->node.owner_document->attrs;
     prefix = attr->node.owner_document->prefix;
 
@@ -641,13 +643,13 @@ lxb_html_tree_adjust_foreign_attributes(lxb_html_tree_t *tree,
 
                 lname_length = adjust->name_len - adjust->prefix_len - 1;
 
-                tag_data = lxb_tag_append_lower(tags,
+                attr_data = lxb_dom_attr_local_name_append(attrs,
                          (const lxb_char_t *) adjust->local_name, lname_length);
-                if (tag_data == NULL) {
+                if (attr_data == NULL) {
                     return LXB_STATUS_ERROR;
                 }
 
-                attr->node.local_name = tag_data->tag_id;
+                attr->node.local_name = attr_data->attr_id;
 
                 prefix_data = lxb_ns_prefix_append(prefix,
                        (const lxb_char_t *) adjust->prefix, adjust->prefix_len);
@@ -1572,7 +1574,7 @@ lxb_html_tree_adoption_agency_algorithm(lxb_html_tree_t *tree,
                 lxb_dom_node_remove(last);
             }
 
-            lxb_dom_node_insert_child(node, last);
+            lxb_dom_node_insert_child_wo_events(node, last);
 
             /* State 14.10 */
             last = node;
@@ -1618,15 +1620,15 @@ lxb_html_tree_adoption_agency_algorithm(lxb_html_tree_t *tree,
             next = node->next;
 
             lxb_dom_node_remove(node);
-            lxb_dom_node_insert_child(lxb_dom_interface_node(element), node);
-
+            lxb_dom_node_insert_child_wo_events(lxb_dom_interface_node(element),
+                                                node);
             node = next;
         }
 
         node = lxb_dom_interface_node(element);
 
         /* State 18 */
-        lxb_dom_node_insert_child(furthest_block, node);
+        lxb_dom_node_insert_child_wo_events(furthest_block, node);
 
         /* State 19 */
         lxb_html_tree_active_formatting_remove(tree, formatting_index);
@@ -1667,7 +1669,7 @@ lxb_html_tree_html_integration_point(lxb_dom_node_t *node)
                                              (const lxb_char_t *) "encoding",
                                              8);
         if (attr == NULL || attr->value == NULL) {
-            return NULL;
+            return false;
         }
 
         if (attr->value->length == 9
