@@ -38,6 +38,11 @@ lxb_selectors_find_by(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
                       lxb_dom_node_t *root, lxb_dom_node_t *node,
                       lxb_css_selector_t *selector, lxb_selectors_cb_f cb, void *ctx);
 
+static lxb_status_t
+lxb_selectors_find_by_reverse(lxb_selectors_t *selectors, lxb_dom_node_t *node,
+                              lxb_css_selector_t *selector,
+                              lxb_selectors_cb_f cb, void *ctx);
+
 static bool
 lxb_selectors_match(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
                     lxb_css_selector_t *selector, lxb_dom_node_t *node);
@@ -85,7 +90,7 @@ lxb_selectors_pseudo_class_read_write(lxb_dom_node_t *node);
 
 static lxb_status_t
 lxb_selectors_first_match(lxb_dom_node_t *node,
-                          lxb_css_selector_specificity_t *spec, void *ctx);
+                          lxb_css_selector_specificity_t spec, void *ctx);
 
 
 lxb_selectors_t *
@@ -205,6 +210,27 @@ lxb_selectors_descendant_next(lxb_selectors_t *selectors, lxb_selectors_entry_t
     while (true);
 }
 
+lxb_inline lxb_dom_node_t *
+lxb_selectors_descendant_reverse(lxb_selectors_t *selectors,
+                                 lxb_selectors_entry_t *entry,
+                                 lxb_css_selector_t *selector,
+                                 lxb_dom_node_t *node)
+{
+    node = node->parent;
+
+    while (node != NULL) {
+        if (node->type == LXB_DOM_NODE_TYPE_ELEMENT &&
+            lxb_selectors_match(selectors, entry, selector, node))
+        {
+            return node;
+        }
+
+        node = node->parent;
+    }
+
+    return false;
+}
+
 lxb_inline lxb_dom_node_t *
 lxb_selectors_close(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
                     lxb_css_selector_t *selector, lxb_dom_node_t *node)
@@ -262,7 +288,7 @@ lxb_selectors_sibling(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
 {
     node = node->next;
 
-    do {
+    while (node != NULL) {
         if (node->type == LXB_DOM_NODE_TYPE_ELEMENT) {
             if (lxb_selectors_match(selectors, entry, selector, node)) {
                 return node;
@@ -273,7 +299,29 @@ lxb_selectors_sibling(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
 
         node = node->next;
     }
-    while (node != NULL);
+
+    return NULL;
+}
+
+lxb_inline lxb_dom_node_t *
+lxb_selectors_sibling_reverse(lxb_selectors_t *selectors,
+                              lxb_selectors_entry_t *entry,
+                              lxb_css_selector_t *selector,
+                              lxb_dom_node_t *node)
+{
+    node = node->prev;
+
+    while (node != NULL) {
+        if (node->type == LXB_DOM_NODE_TYPE_ELEMENT) {
+            if (lxb_selectors_match(selectors, entry, selector, node)) {
+                return node;
+            }
+
+            return NULL;
+        }
+
+        node = node->prev;
+    }
 
     return NULL;
 }
@@ -297,13 +345,34 @@ lxb_selectors_following(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry
     return NULL;
 }
 
+lxb_inline lxb_dom_node_t *
+lxb_selectors_following_reverse(lxb_selectors_t *selectors,
+                                lxb_selectors_entry_t *entry,
+                                lxb_css_selector_t *selector,
+                                lxb_dom_node_t *node)
+{
+    node = node->prev;
+
+    while (node != NULL) {
+        if (node->type == LXB_DOM_NODE_TYPE_ELEMENT &&
+            lxb_selectors_match(selectors, entry, selector, node))
+        {
+            return node;
+        }
+
+        node = node->prev;
+    }
+
+    return NULL;
+}
+
 lxb_inline bool
 lxb_selectors_anb_calc(lxb_css_selector_anb_of_t *anb, size_t index)
 {
     double num;
 
     if (anb->anb.a == 0) {
-        if (anb->anb.b >= 0 && anb->anb.b == index) {
+        if (anb->anb.b >= 0 && (size_t) anb->anb.b == index) {
             return true;
         }
     }
@@ -339,6 +408,28 @@ lxb_selectors_find(lxb_selectors_t *selectors, lxb_dom_node_t *root,
     return LXB_STATUS_OK;
 }
 
+lxb_status_t
+lxb_selectors_find_reverse(lxb_selectors_t *selectors, lxb_dom_node_t *root,
+                           lxb_css_selector_list_t *list,
+                           lxb_selectors_cb_f cb, void *ctx)
+{
+    lxb_status_t status;
+
+    while (list != NULL) {
+        status = lxb_selectors_find_by_reverse(selectors, root,
+                                               list->last, cb, ctx);
+        if (status != LXB_STATUS_OK) {
+            return status;
+        }
+
+        list = list->next;
+    }
+
+    lxb_selectors_clean(selectors);
+
+    return LXB_STATUS_OK;
+}
+
 static lxb_selectors_entry_t *
 lxb_selectors_find_by_selector(lxb_selectors_t *selectors, lxb_dom_node_t *root,
                                lxb_selectors_entry_t *entry,
@@ -590,7 +681,7 @@ lxb_selectors_find_by(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
         }
 
         if (selector->next == NULL) {
-            status = cb(node, NULL, ctx);
+            status = cb(node, selector->list->specificity, ctx);
             if (status != LXB_STATUS_OK) {
                 if (status == LXB_STATUS_STOP) {
                     return LXB_STATUS_OK;
@@ -702,6 +793,91 @@ lxb_selectors_find_by(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
     return LXB_STATUS_OK;
 }
 
+static lxb_status_t
+lxb_selectors_find_by_reverse(lxb_selectors_t *selectors, lxb_dom_node_t *node,
+                              lxb_css_selector_t *selector,
+                              lxb_selectors_cb_f cb, void *ctx)
+{
+    lxb_status_t status;
+    lxb_dom_node_t *base = node;
+    lxb_selectors_entry_t *entry;
+    lxb_css_selector_combinator_t combinator;
+
+    entry = lexbor_dobject_calloc(selectors->objs);
+    if (entry == NULL) {
+        return LXB_STATUS_ERROR_MEMORY_ALLOCATION;
+    }
+
+    entry->selector = selector;
+    combinator = LXB_CSS_SELECTOR_COMBINATOR_CLOSE;
+
+    do {
+        entry->node = node;
+        entry->id = 0;
+
+        selector = entry->selector;
+
+        switch (combinator) {
+            case LXB_CSS_SELECTOR_COMBINATOR_DESCENDANT:
+                node = lxb_selectors_descendant_reverse(selectors, entry,
+                                                        selector, node);
+                break;
+
+            case LXB_CSS_SELECTOR_COMBINATOR_CLOSE:
+                node = lxb_selectors_close(selectors, entry, selector, node);
+                break;
+
+            case LXB_CSS_SELECTOR_COMBINATOR_CHILD:
+                node = node->parent;
+
+                if (node == NULL || node->type != LXB_DOM_NODE_TYPE_ELEMENT
+                    || !lxb_selectors_match(selectors, entry, selector, node))
+                {
+                    return LXB_STATUS_OK;
+                }
+
+                break;
+
+            case LXB_CSS_SELECTOR_COMBINATOR_SIBLING:
+                node = lxb_selectors_sibling_reverse(selectors, entry,
+                                                     selector, node);
+                break;
+
+            case LXB_CSS_SELECTOR_COMBINATOR_FOLLOWING:
+                node = lxb_selectors_following_reverse(selectors, entry,
+                                                       selector, node);
+                break;
+
+            case LXB_CSS_SELECTOR_COMBINATOR_CELL:
+            default:
+                return LXB_STATUS_ERROR;
+        }
+
+        if (node == NULL) {
+            return LXB_STATUS_OK;
+        }
+
+        if (selector->prev == NULL) {
+            status = cb(base, selector->list->specificity, ctx);
+            if (status != LXB_STATUS_OK) {
+                if (status == LXB_STATUS_STOP) {
+                    return LXB_STATUS_OK;
+                }
+
+                return status;
+            }
+
+            return LXB_STATUS_OK;
+        }
+
+        combinator = selector->combinator;
+        entry->selector = selector->prev;
+    }
+    while (true);
+
+    return LXB_STATUS_OK;
+}
+
 static bool
 lxb_selectors_match(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
                     lxb_css_selector_t *selector, lxb_dom_node_t *node)
@@ -743,7 +919,7 @@ lxb_selectors_match(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
         case LXB_CSS_SELECTOR_TYPE_ID:
             element = lxb_dom_interface_element(node);
 
-            if (element->attr_id == NULL) {
+            if (element->attr_id == NULL || element->attr_id->value == NULL) {
                 return false;
             }
 
@@ -761,7 +937,7 @@ lxb_selectors_match(lxb_selectors_t *selectors, lxb_selectors_entry_t *entry,
         case LXB_CSS_SELECTOR_TYPE_CLASS:
             element = lxb_dom_interface_element(node);
 
-            if (element->attr_class == NULL) {
+            if (element->attr_class == NULL || element->attr_class->value == NULL) {
                 return false;
             }
 
@@ -931,7 +1107,7 @@ lxb_selectors_match_class(const lexbor_str_t *target, const  lexbor_str_t *src,
 
         if (lexbor_utils_whitespace(chr, ==, ||)) {
 
-            if ((data - pos) == src->length) {
+            if ((size_t) (data - pos) == src->length) {
                 if (quirks) {
                     is_it = lexbor_str_data_ncasecmp(pos, src->data, src->length);
                 }
@@ -944,7 +1120,7 @@ lxb_selectors_match_class(const lexbor_str_t *target, const  lexbor_str_t *src,
                 }
             }
 
-            if ((end - data) < src->length) {
+            if ((size_t) (end - data) < src->length) {
                 return false;
             }
 
@@ -952,7 +1128,7 @@ lxb_selectors_match_class(const lexbor_str_t *target, const  lexbor_str_t *src,
         }
     }
 
-    if ((end - pos) == src->length && src->length != 0) {
+    if ((size_t) (end - pos) == src->length && src->length != 0) {
         if (quirks) {
             is_it = lexbor_str_data_ncasecmp(pos, src->data, src->length);
         }
@@ -1584,7 +1760,7 @@ lxb_selectors_pseudo_class_read_write(lxb_dom_node_t *node)
 
 static lxb_status_t
 lxb_selectors_first_match(lxb_dom_node_t *node,
-                          lxb_css_selector_specificity_t *spec, void *ctx)
+                          lxb_css_selector_specificity_t spec, void *ctx)
 {
     *((bool *) ctx) = true;
     return LXB_STATUS_STOP;
